// $Id: LOCA_Continuation_HouseholderGroup.C,v 1.1 2004/04/27 19:37:12 etphipp Exp $
// $Source: /space/CVS/Trilinos/packages/nox/src-loca/src/LOCA_Continuation_HouseholderGroup.C,v $

//@HEADER
// ************************************************************************
// 
//            NOX: An Object-Oriented Nonlinear Solver Package
//                 Copyright (2002) Sandia Corporation
// 
// Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
// 
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
//  
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                                 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA                                                                                
// Questions? Contact Tammy Kolda (tgkolda@sandia.gov) or Roger Pawlowski
// (rppawlo@sandia.gov), Sandia National Laboratories.
// 
// ************************************************************************
//@HEADER

#include "LOCA_Continuation_HouseholderGroup.H"
#include "LOCA_ErrorCheck.H"
#include "LOCA_Utils.H"

LOCA::Continuation::HouseholderGroup::HouseholderGroup(
				 LOCA::Continuation::AbstractGroup& g,
				 int paramID,
				 NOX::Parameter::List& params)
  : LOCA::Continuation::ArcLengthGroup(g, paramID, params), 
    houseVec(g.getX(), 0.0),
    beta(0.0)
{
}

LOCA::Continuation::HouseholderGroup::HouseholderGroup(
				  LOCA::Continuation::AbstractGroup& g,
				  string paramID,
				  NOX::Parameter::List& params)
  : LOCA::Continuation::ArcLengthGroup(g, paramID, params), 
    houseVec(g.getX(), 0.0),
    beta(0.0)
{
}

LOCA::Continuation::HouseholderGroup::HouseholderGroup(
		   const LOCA::Continuation::HouseholderGroup& source,
		   NOX::CopyType type)
  :  LOCA::Continuation::ArcLengthGroup(source, type), 
     houseVec(source.houseVec, type),
     beta(source.beta)
{
}


LOCA::Continuation::HouseholderGroup::~HouseholderGroup() 
{
}

LOCA::Continuation::ArcLengthGroup&
LOCA::Continuation::HouseholderGroup::operator=(
			      const LOCA::Continuation::ArcLengthGroup& source)
{
  return *this = 
    dynamic_cast<const LOCA::Continuation::HouseholderGroup&>(source);
}

LOCA::Continuation::ExtendedGroup&
LOCA::Continuation::HouseholderGroup::operator=(
			      const LOCA::Continuation::ExtendedGroup& source)
{
  return *this = 
    dynamic_cast<const LOCA::Continuation::HouseholderGroup&>(source);
}

NOX::Abstract::Group&
LOCA::Continuation::HouseholderGroup::operator=(
					  const NOX::Abstract::Group& source)
{
  return *this = 
    dynamic_cast<const LOCA::Continuation::HouseholderGroup&>(source);
}

LOCA::Continuation::HouseholderGroup&
LOCA::Continuation::HouseholderGroup::operator=(
		   const LOCA::Continuation::HouseholderGroup& source) 
{

  // Protect against A = A
  if (this != &source) {
    LOCA::Continuation::ArcLengthGroup::operator=(source);

    // Copy values
    houseVec = source.houseVec;
    beta = source.beta;
  }

  return *this;
}

NOX::Abstract::Group*
LOCA::Continuation::HouseholderGroup::clone(NOX::CopyType type) const 
{
  return new LOCA::Continuation::HouseholderGroup(*this);
}

NOX::Abstract::Group::ReturnType
LOCA::Continuation::HouseholderGroup::computeF() 
{
  // Call computeF() of arclength group
  NOX::Abstract::Group::ReturnType finalStatus = 
    LOCA::Continuation::ArcLengthGroup::computeF();
  
  // Set parameter component to zero
  fVec.getParam() = 0.0;

  return finalStatus;
}

NOX::Abstract::Group::ReturnType
LOCA::Continuation::HouseholderGroup::computeNewton(
					       NOX::Parameter::List& params) 
{
  if (isValidNewton)
    return NOX::Abstract::Group::Ok;

  string callingFunction = 
    "LOCA::Continuation::HouseholderGroup::computeNewton()";
  NOX::Abstract::Group::ReturnType finalStatus = NOX::Abstract::Group::Ok;
  NOX::Abstract::Group::ReturnType status;

  // Make sure F is valid
  if (!isF()) {
    status = computeF();
    finalStatus = 
      LOCA::ErrorCheck::combineAndCheckReturnTypes(status, finalStatus,
						   callingFunction);
  }
  
  // Make sure Jacobian is valid
  if (!isJacobian()) {
    status = computeJacobian();
    finalStatus = 
      LOCA::ErrorCheck::combineAndCheckReturnTypes(status, finalStatus,
						   callingFunction);
  }

  // Compute Householder vector
  computeHouseholderVector();

  // zero out newton vec -- used as initial guess for some linear solvers
  newtonVec.init(0.0);

  // Get x component of fVec, newtonVec, houseVec
  const NOX::Abstract::Vector& fVec_x = fVec.getXVec();
  const NOX::Abstract::Vector& houseVec_x = houseVec.getXVec();
  NOX::Abstract::Vector& newtonVec_x = newtonVec.getXVec();
  
  // Get param component of houseVec, newtonVec
  const double& houseVec_p = houseVec.getParam();
  double& newtonVec_p = newtonVec.getParam();

  // Solve continuation equations using Householder projection
  status = grpPtr->applyHouseholderJacobianInverse(params, fVec_x, 
						   *derivResidualParamPtr,
						   houseVec_x, houseVec_p, 
						   beta, newtonVec_x,
						   newtonVec_p);
  finalStatus = 
    LOCA::ErrorCheck::combineAndCheckReturnTypes(status, finalStatus,
						 callingFunction);

  // Change sign of newton vector
  newtonVec.scale(-1.0);

  //double r = computeScaledDotProduct(newtonVec, predictorVec);
  double r = newtonVec.dot(predictorVec);
  cout << "Component of newton vec in direction of predictor:  " 
       << r << endl;

  isValidNewton = true;

  return finalStatus;
}

void
LOCA::Continuation::HouseholderGroup::computeHouseholderVector()
{
  
  // Get x, param components of predictor vector
  const NOX::Abstract::Vector& tanX = predictorVec.getXVec();
  const double& tanP = predictorVec.getParam();

  // Get x, param components of Householder vector
  NOX::Abstract::Vector& houseX = houseVec.getXVec();
  double& houseP = houseVec.getParam();

  double sigma = tanX.dot(tanX);
  houseX = tanX;

  if (sigma == 0.0) {
    beta = 0.0;
    houseP = 1.0;
  }
  else {
    double mu = sqrt(tanP*tanP + sigma);
    if (tanP <= 0.0)
      houseP = tanP - mu;
    else
      houseP = -sigma / (tanP + mu);
    beta = 2.0*houseP*houseP/(sigma + houseP*houseP);
    houseVec.scale(1.0/houseP);
  }

  return;
}



